import json
import tempfile
import subprocess
from PySide2 import QtCore, QtGui, QtWidgets
from anytree import NodeMixin
from anytree.exporter import DotExporter
from .slot import ConnectionItem


class NodeScene(QtWidgets.QGraphicsScene):
    signal_NodeMoved = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.nodes = dict()

    def drawBackground(self, painter, rect):
        config = self.parent().config
        self._brush = QtGui.QBrush()
        self._brush.setStyle(QtCore.Qt.SolidPattern)
        self._brush.setColor(QtGui.QColor(*config['bg_color']))
        painter.fillRect(rect, self._brush)

    def updateScene(self):
        for item in self.items():
            if isinstance(item, ConnectionItem):
                item.target_point = item.target.center()
                item.source_point = item.source.center()
                item.updatePath()

    def setNodes(self, root):
        assert isinstance(root, NodeMixin)

        self.root = root
        view = self.views()[0]
        nodes = [root] + list(root.descendants)
        for n_data in nodes:
            view.createNode(n_data)

        for n_data in nodes:
            if not n_data.is_root:
                view.createConnection(n_data.parent, n_data)

        self.setFormalPosition()

    def setFormalPosition(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            out_path = tmpdir + '/tree.dot'
            DotExporter(self.root,
                        nodenamefunc=lambda n: id(n),
                        ).to_dotfile(out_path)
            cmd = f'dot {out_path} -T json0'.split()
            res = subprocess.run(cmd, capture_output=True)

        objects = json.loads(res.stdout.decode())['objects']
        y_max = max([float(obj['pos'].split(',')[1]) for obj in objects])
        for obj in objects:
            pos = obj['pos'].split(',')
            x = float(pos[0]) * 5 + 1e+3
            y = (y_max - float(pos[1])) * 5 + 1e+3
            nodeItem = self.nodes[int(obj['name'])]
            position = QtCore.QPointF(x, y)
            nodeItem.setPos(position - nodeItem.nodeCenter)

            connections = nodeItem.slot_parent.connections.copy()
            connections += nodeItem.slot_child.connections
            for connection in connections:
                slot = connection.childSlotItem
                connection.source_point = slot.center()
                slot = connection.parentSlotItem
                connection.target_point = slot.center()
                connection.updatePath()



from PySide2 import QtCore


def _createPointerBoundingBox(pointerPos, bbSize):
    """
    generate a bounding box around the pointer.

    :param pointerPos: Pointer position.
    :type  pointerPos: QPoint.

    :param bbSize: Width and Height of the bounding box.
    :type  bbSize: Int.

    """
    # Create pointer's bounding box.
    point = pointerPos

    mbbPos = point
    point.setX(point.x() - bbSize / 2)
    point.setY(point.y() - bbSize / 2)

    size = QtCore.QSize(bbSize, bbSize)
    bb = QtCore.QRect(mbbPos, size)
    bb = QtCore.QRectF(bb)

    return bb



from PySide2 import QtCore, QtGui, QtWidgets
from .slot import SlotItem, ConnectionItem


class NodeItem(QtWidgets.QGraphicsItem):
    def __init__(self, view, data):
        super().__init__()

        self.setZValue(1)

        # Storage
        self.data = data
        self.slot_child = None
        self.slot_parent = None

        # Methods.
        self._createStyle(view.config)

        self.image = getattr(data, 'image', None)
        if self.image is not None:
            from PIL.ImageQt import ImageQt
            img_qt = ImageQt(self.image)
            self.image = QtGui.QPixmap.fromImage(img_qt)

        self.dialog = None
        if view.CustomDialog is not None:
            self.dialog = view.CustomDialog(data, parent=view)
            self.dialog.accepted.connect(view.signal_DialogAccepted)

    @property
    def name(self):
        return self.data.name

    @property
    def pen(self):
        if self.isSelected():
            return self._penSel
        else:
            return self._pen

    def _createStyle(self, config):
        self.setAcceptHoverEvents(True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable)

        # Dimensions.
        self.baseWidth = config['node']['width']
        self.baseHeight = config['node']['height']
        self.border = config['node']['border']
        self.radius = config['node']['radius']

        self.height = (self.baseHeight +
                       self.border +
                       0.5 * self.radius)

        self.nodeCenter = QtCore.QPointF()
        self.nodeCenter.setX(self.baseWidth / 2.0)
        self.nodeCenter.setY(self.height / 2.0)

        self._brush = QtGui.QBrush()
        self._brush.setStyle(QtCore.Qt.SolidPattern)
        self._brush.setColor(QtGui.QColor(*config['node']['color']['main']))

        self._pen = QtGui.QPen()
        self._pen.setStyle(QtCore.Qt.SolidLine)
        self._pen.setWidth(self.border)
        self._pen.setColor(QtGui.QColor(*config['node']['color']['border']))

        self._penSel = QtGui.QPen()
        self._penSel.setStyle(QtCore.Qt.SolidLine)
        self._penSel.setWidth(self.border)
        self._penSel.setColor(
            QtGui.QColor(*config['node']['color']['border_sel'])
        )

        self._textPen = QtGui.QPen()
        self._textPen.setStyle(QtCore.Qt.SolidLine)
        self._textPen.setColor(QtGui.QColor(*config['text_color']))

        self._rootTextFont = QtGui.QFont(
            config['font'],
            config['root_font_size'],
            QtGui.QFont.Bold
        )
        self._nodeTextFont = QtGui.QFont(
            config['font'],
            config['node_font_size'],
            QtGui.QFont.Bold
        )

    def _createSlot(self, slot_parent=True, slot_child=True):
        if slot_parent:
            self.slot_parent = SlotItem(parent=self, slot_type='parent')

        if slot_child:
            self.slot_child = SlotItem(parent=self, slot_type='child')

    def _remove(self):
        self.scene().nodes.pop(id(self.data))

        # Remove all parent connections.
        if self.slot_parent is not None:
            connections = self.slot_parent.connections
            while len(connections) > 0:
                connections[0]._remove()

        # Remove all child connections.
        if self.slot_child is not None:
            connections = self.slot_child.connections
            while len(connections) > 0:
                connections[0]._remove()

        # Remove node.
        scene = self.scene()
        scene.removeItem(self)
        scene.update()

    def boundingRect(self):
        rect = QtCore.QRect(0, 0, self.baseWidth, self.height)
        rect = QtCore.QRectF(rect)
        return rect

    def shape(self):
        path = QtGui.QPainterPath()
        path.addRect(self.boundingRect())
        return path

    def paint(self, painter, option, widget):
        # Node base.
        painter.setBrush(self._brush)
        painter.setPen(self.pen)

        painter.drawRoundedRect(0, 0,
                                self.baseWidth,
                                self.height,
                                self.radius,
                                self.radius)

        # Root label
        if self.scene().root == self.data:
            label = 'Root'
            painter.setPen(self._textPen)
            painter.setFont(self._rootTextFont)

            metrics = QtGui.QFontMetrics(painter.font())
            text_width = metrics.boundingRect(label).width() + 14
            text_height = metrics.boundingRect(label).height() + 32
            margin = (self.baseWidth - text_width) * 0.5
            textRect = QtCore.QRect(margin,
                                    -text_height,
                                    text_width,
                                    text_height)

            painter.drawText(textRect,
                             QtCore.Qt.AlignCenter,
                             label)

        if self.image is None:
            painter.setPen(self._textPen)
            painter.setFont(self._nodeTextFont)

            metrics = QtGui.QFontMetrics(painter.font())
            rect = metrics.boundingRect(self.name)
            text_width = rect.width() + 14
            text_height = rect.height() + 14
            x = (self.baseWidth - text_width) * 0.5
            y = (self.height - text_height) * 0.5
            textRect = QtCore.QRect(x, y,
                                    text_width,
                                    text_height)

            painter.drawText(textRect,
                             QtCore.Qt.AlignCenter,
                             self.name)
        else:
            img = self.image.scaled(self.baseWidth - self.border * 2,
                                    self.height - self.border * 2, QtCore.Qt.KeepAspectRatio)
            x = (self.baseWidth - img.width()) / 2
            y = (self.height - img.height()) / 2
            painter.drawPixmap(x, y, img)

    def mousePressEvent(self, event):
        nodes = self.scene().nodes
        for node in nodes.values():
            node.setZValue(1)

        for item in self.scene().items():
            if isinstance(item, ConnectionItem):
                item.setZValue(1)

        self.setZValue(2)

        super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event):
        if self.dialog is not None:
            self.dialog.exec_()
        super().mouseDoubleClickEvent(event)
        self.scene().parent().signal_NodeDoubleClicked.emit(self.name)

    def mouseMoveEvent(self, event):
        self.scene().updateScene()
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        self.scene().signal_NodeMoved.emit(self.name, self.pos())
        super().mouseReleaseEvent(event)

    def hoverLeaveEvent(self, event):
        view = self.scene().views()[0]

        for item in view.scene().items():
            if isinstance(item, ConnectionItem):
                item.setZValue(0)

        super().hoverLeaveEvent(event)



from PySide2 import QtCore, QtGui, QtWidgets
from . import util


class SlotItem(QtWidgets.QGraphicsItem):
    def __init__(self, parent, slot_type):
        super().__init__(parent)

        # Status.
        self.setAcceptHoverEvents(True)

        # Storage.
        self.slotType = slot_type

        # Style.
        self._createStyle(parent)

        self.pen = QtGui.QPen()
        self.pen.setStyle(QtCore.Qt.SolidLine)

        self.setFlag(SlotItem.ItemStacksBehindParent)

        # Connections storage.
        self.connected_slots = list()
        self.newConnection = None
        self.connections = list()
        if slot_type == 'parent':
            self.maxConnections = 1
        elif slot_type == 'child':
            self.maxConnections = -1
        else:
            raise RuntimeError

    def accepts(self, slot_item):
        # no child on child or parent on parent
        hasChildItem = 'child' in [self.slotType, slot_item.slotType]
        hasParentItem = 'parent' in [self.slotType, slot_item.slotType]
        if not (hasChildItem and hasParentItem):
            return False

        # no self connection
        if self.parentItem() == slot_item.parentItem():
            return False

        # no more than maxConnections
        if self.maxConnections > 0 and len(self.connected_slots) >= self.maxConnections:
            return False

        # no loop
        if self.__check_no_loop(slot_item):
            return False

        # otherwize, all fine.
        return True

    def __check_no_loop(self, slot_item):
        if self.slotType == 'child':
            parent = self.parentItem().data
        elif self.slotType == 'parent':
            child = self.parentItem().data

        if slot_item.slotType == 'child':
            parent = slot_item.parentItem().data
        elif slot_item.slotType == 'parent':
            child = slot_item.parentItem().data

        return any(ancestor is child
                   for ancestor in parent.iter_path_reverse())

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.newConnection = ConnectionItem(self.center(),
                                                self.mapToScene(event.pos()),
                                                self,
                                                None)

            self.connections.append(self.newConnection)
            self.scene().addItem(self.newConnection)

            view = self.scene().views()[0]
            view.drawingConnection = True
            view.sourceSlot = self
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        from .node import NodeItem

        view = self.scene().views()[0]
        config = view.config
        if view.drawingConnection:
            mbb = util._createPointerBoundingBox(pointerPos=event.scenePos().toPoint(),
                                                 bbSize=config['mouse_bounding_box'])

            # Get nodes in pointer's bounding box.
            targets = self.scene().items(mbb)

            if any(isinstance(target, NodeItem) for target in targets):
                if self.parentItem() not in targets:
                    for target in targets:
                        if isinstance(target, NodeItem):
                            view.currentHoveredNode = target
            else:
                view.currentHoveredNode = None

            # Set connection's end point.
            self.newConnection.target_point = self.mapToScene(event.pos())
            self.newConnection.updatePath()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        view = self.scene().views()[0]
        if event.button() == QtCore.Qt.LeftButton:
            view.drawingConnection = False

            target = self.scene().itemAt(event.scenePos().toPoint(), QtGui.QTransform())

            if not isinstance(target, SlotItem):
                self.newConnection._remove()
                super().mouseReleaseEvent(event)
                return

            if target.accepts(self):
                self.newConnection.target = target
                self.newConnection.source = self
                self.newConnection.target_point = target.center()
                self.newConnection.source_point = self.center()

                # Perform the ConnectionItem.
                self.connect(target, self.newConnection)
                target.connect(self, self.newConnection)

                self.newConnection.updatePath()
            else:
                self.newConnection._remove()
        else:
            super().mouseReleaseEvent(event)

        view.currentHoveredNode = None

    def shape(self):
        path = QtGui.QPainterPath()
        path.addRect(self.boundingRect())
        return path

    def paint(self, painter, option, widget):
        painter.setBrush(self.brush)
        painter.setPen(self.pen)

        view = self.scene().views()[0]
        config = view.config
        if view.drawingConnection:
            if self.parentItem() == view.currentHoveredNode:
                painter.setBrush(
                    QtGui.QColor(*config['connection']
                                 ['color']['non_connectable'])
                )
                if self.slotType != view.sourceSlot.slotType:
                    if not view.sourceSlot.__check_no_loop(self):
                        _penValid = QtGui.QPen()
                        _penValid.setStyle(QtCore.Qt.SolidLine)
                        _penValid.setWidth(2)
                        _penValid.setColor(QtGui.QColor(255, 255, 255, 255))
                        painter.setPen(_penValid)
                        painter.setBrush(self.brush)

        painter.drawEllipse(self.boundingRectForPaint())

    def center(self):
        rect = self.boundingRect()
        center = QtCore.QPointF(rect.x() + rect.width() * 0.5,
                                rect.y() + rect.height() * 0.5)

        return self.mapToScene(center)

    def _createStyle(self, parent):
        config = parent.scene().views()[0].config
        self.brush = QtGui.QBrush()
        self.brush.setStyle(QtCore.Qt.SolidPattern)
        self.brush.setColor(QtGui.QColor(*config['slot']['color']['main']))

    def boundingRect(self):
        width = height = self.parentItem().baseWidth / 2.0
        x = (self.parentItem().baseWidth - width) / 2.
        y = - height / 2.0
        if self.slotType == 'child':
            y += self.parentItem().height
        rect = QtCore.QRectF(QtCore.QRect(x, y, width, height))
        return rect

    def boundingRectForPaint(self):
        width = height = self.parentItem().baseWidth / 10.0
        x = (self.parentItem().baseWidth - width) / 2.
        y = - height / 2.0
        if self.slotType == 'child':
            y += self.parentItem().height
        rect = QtCore.QRectF(QtCore.QRect(x, y, width, height))
        return rect

    def connect(self, slot_item, connection):
        scene = self.scene()
        view = scene.views()[0]

        if self.maxConnections > 0 and len(self.connected_slots) >= self.maxConnections:
            # Already connected.
            self.connections[self.maxConnections - 1]._remove()

        # Populate connection.
        if slot_item.slotType == 'parent':
            connection.parentSlotItem = slot_item
            connection.childNode = slot_item.parentItem()
            connection.childSlotItem = self
            connection.parentNode = self.parentItem()
        else:
            connection.childSlotItem = slot_item
            connection.parentNode = slot_item.parentItem()
            connection.parentSlotItem = self
            connection.childNode = self.parentItem()

        # Add slot to connected slots.
        if slot_item not in self.connected_slots:
            self.connected_slots.append(slot_item)

        # Add connection.
        if connection not in self.connections:
            self.connections.append(connection)

        # Connect data
        child = connection.childNode
        parent = connection.parentNode
        if child is not None and parent is not None:
            child = child.data
            parent = parent.data
            child.parent = parent
            if scene.root != parent.root:
                scene.root = parent.root
                view.signal_RootUpdated.emit(parent.root)

        # Emit signal.
        view.signal_Connected.emit(
            connection.childNode, connection.parentNode
        )

    def disconnect(self, connection):
        view = self.scene().views()[0]

        # Remove slot from connected slots
        if self.slotType == 'parent':
            slot_item = connection.childSlotItem
        else:
            slot_item = connection.parentSlotItem

        if slot_item in self.connected_slots:
            self.connected_slots.remove(slot_item)

        # Remove connections
        self.connections.remove(connection)

        # Disconnect data
        child = connection.childNode
        parent = connection.parentNode
        if child is not None and parent is not None:
            child.data.parent = None

        # Emit signal.
        view.signal_Disconnected.emit(
            connection.childNode, connection.parentNode
        )


class ConnectionItem(QtWidgets.QGraphicsPathItem):
    def __init__(self, source_point, target_point, source, target):
        super().__init__()

        self.setZValue(1)

        # Storage.
        self.parentNode = None
        self.childNode = None

        self.source_point = source_point
        self.target_point = target_point
        self.source = source
        self.target = target

        self.childSlotItem = None
        self.parentSlotItem = None

        self.movable_point = None

        self.data = tuple()

        # Methods.
        self._createStyle()

    def _createStyle(self):
        config = self.source.scene().views()[0].config
        self.setAcceptHoverEvents(True)
        self.setZValue(-1)

        self._pen = QtGui.QPen(
            QtGui.QColor(*config['connection']['color']['main'])
        )
        self._pen.setWidth(config['connection']['width'])

    def mousePressEvent(self, event):
        view = self.scene().views()[0]

        for item in view.scene().items():
            if isinstance(item, ConnectionItem):
                item.setZValue(0)

        view.drawingConnection = True

        d_to_target = (event.pos() - self.target_point).manhattanLength()
        d_to_source = (event.pos() - self.source_point).manhattanLength()
        if d_to_target < d_to_source:
            self.target_point = event.pos()
            self.movable_point = 'target_point'
            self.target.disconnect(self)
            self.target = None
            view.sourceSlot = self.source
        else:
            self.source_point = event.pos()
            self.movable_point = 'source_point'
            self.source.disconnect(self)
            self.source = None
            view.sourceSlot = self.target

        self.updatePath()

    def mouseMoveEvent(self, event):
        from .node import NodeItem
        view = self.scene().views()[0]
        config = view.config

        mbb = util._createPointerBoundingBox(pointerPos=event.scenePos().toPoint(),
                                             bbSize=config['mouse_bounding_box'])

        # Get nodes in pointer's bounding box.
        targets = self.scene().items(mbb)

        if any(isinstance(target, NodeItem) for target in targets):

            if view.sourceSlot.parentItem() not in targets:
                for target in targets:
                    if isinstance(target, NodeItem):
                        view.currentHoveredNode = target
        else:
            view.currentHoveredNode = None

        if self.movable_point == 'target_point':
            self.target_point = event.pos()
        else:
            self.source_point = event.pos()

        self.updatePath()

    def mouseReleaseEvent(self, event):
        view = self.scene().views()[0]
        view.drawingConnection = False

        slot = self.scene().itemAt(event.scenePos().toPoint(), QtGui.QTransform())

        if not isinstance(slot, SlotItem):
            self._remove()
            self.updatePath()
            super().mouseReleaseEvent(event)
            return

        if self.movable_point == 'target_point':
            if slot.accepts(self.source):
                self.target = slot
                self.target_point = slot.center()
                slot_child = self.source
                slot_parent = self.target
                slot_parent.connect(slot_child, self)
                self.updatePath()
            else:
                self._remove()

        else:
            if slot.accepts(self.target):
                self.source = slot
                self.source_point = slot.center()
                slot_parent = self.target
                slot_child = self.source
                slot_child.connect(slot_parent, self)
                self.updatePath()
            else:
                self._remove()

    def _remove(self):
        if self.source is not None:
            self.source.disconnect(self)
        if self.target is not None:
            self.target.disconnect(self)

        scene = self.scene()
        scene.removeItem(self)
        scene.update()

    def updatePath(self):
        self.setPen(self._pen)

        path = QtGui.QPainterPath()
        path.moveTo(self.source_point)
        dx = (self.target_point.x() - self.source_point.x()) * 0.5
        dy = self.target_point.y() - self.source_point.y()
        ctrl1 = QtCore.QPointF(self.source_point.x() + dx,
                               self.source_point.y() + dy * 0)
        ctrl2 = QtCore.QPointF(self.source_point.x() + dx,
                               self.source_point.y() + dy * 1)
        path.cubicTo(ctrl1, ctrl2, self.target_point)

        self.setPath(path)



import json
from pathlib import Path
from anytree import NodeMixin
from PySide2 import QtGui, QtCore, QtWidgets
from .scene import NodeScene
from .node import NodeItem
from .slot import ConnectionItem

defaultConfigPath = Path(__file__).with_name('default_config.json')


class NodeView(QtWidgets.QGraphicsView):
    signal_NodeCreated = QtCore.Signal(object)
    signal_NodeDeleted = QtCore.Signal(object)
    signal_NodeSelected = QtCore.Signal(object)
    signal_NodeMoved = QtCore.Signal(str, object)
    signal_NodeDoubleClicked = QtCore.Signal(str)

    signal_RootUpdated = QtCore.Signal(object)
    signal_Connected = QtCore.Signal(object, object)
    signal_Disconnected = QtCore.Signal(object, object)
    signal_DialogAccepted = QtCore.Signal()

    signal_KeyPressed = QtCore.Signal(object)
    signal_Dropped = QtCore.Signal()

    def __init__(self, root, CustomDialog=None, parent=None,
                 configPath=defaultConfigPath):
        super().__init__(parent)

        # Load configuration.
        self.loadConfig(configPath)
        self.CustomDialog = CustomDialog

        # General data.
        self.selectedNodes = None

        # Connections data.
        self.drawingConnection = False
        self.currentHoveredNode = None
        self.sourceSlot = None

        # Display options.
        self.currentState = 'DEFAULT'
        self.pressedKeys = list()

        self.initialize(root)
        self._focus()

    def wheelEvent(self, event):
        self.currentState = 'ZOOM_VIEW'
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)

        inFactor = 1.15
        outFactor = 1 / inFactor

        if event.delta() > 0:
            zoomFactor = inFactor
        else:
            zoomFactor = outFactor

        self.scale(zoomFactor, zoomFactor)
        self.currentState = 'DEFAULT'

    def mousePressEvent(self, event):
        # Tablet zoom
        if (event.button() == QtCore.Qt.RightButton and
                event.modifiers() == QtCore.Qt.AltModifier):
            self.currentState = 'ZOOM_VIEW'
            self.initMousePos = event.pos()
            self.zoomInitialPos = event.pos()
            self.initMouse = QtGui.QCursor.pos()
            self.setInteractive(False)

        # Drag view
        elif (event.button() == QtCore.Qt.MiddleButton and
              event.modifiers() == QtCore.Qt.AltModifier):
            self.currentState = 'DRAG_VIEW'
            self.prevPos = event.pos()
            self.setCursor(QtCore.Qt.ClosedHandCursor)
            self.setInteractive(False)

        # Rubber band selection
        elif (event.button() == QtCore.Qt.LeftButton and
              event.modifiers() == QtCore.Qt.NoModifier and
              self.scene().itemAt(self.mapToScene(event.pos()), QtGui.QTransform()) is None):
            self.currentState = 'SELECTION'
            self._initRubberband(event.pos())
            self.setInteractive(False)

        # Drag Item
        elif (event.button() == QtCore.Qt.LeftButton and
              event.modifiers() == QtCore.Qt.NoModifier and
              self.scene().itemAt(self.mapToScene(event.pos()), QtGui.QTransform()) is not None):
            self.currentState = 'DRAG_ITEM'
            self.setInteractive(True)

        # Add selection
        elif (event.button() == QtCore.Qt.LeftButton and
              QtCore.Qt.Key_Shift in self.pressedKeys and
              QtCore.Qt.Key_Control in self.pressedKeys):
            self.currentState = 'ADD_SELECTION'
            self._initRubberband(event.pos())
            self.setInteractive(False)

        # Subtract selection
        elif (event.button() == QtCore.Qt.LeftButton and
              event.modifiers() == QtCore.Qt.ControlModifier):
            self.currentState = 'SUBTRACT_SELECTION'
            self._initRubberband(event.pos())
            self.setInteractive(False)

        # Toggle selection
        elif (event.button() == QtCore.Qt.LeftButton and
              event.modifiers() == QtCore.Qt.ShiftModifier):
            self.currentState = 'TOGGLE_SELECTION'
            self._initRubberband(event.pos())
            self.setInteractive(False)

        else:
            self.currentState = 'DEFAULT'

        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        # Zoom.
        if self.currentState == 'ZOOM_VIEW':
            offset = self.zoomInitialPos.x() - event.pos().x()

            if offset > self.previousMouseOffset:
                self.previousMouseOffset = offset
                self.zoomDirection = -1
                self.zoomIncr -= 1

            elif offset == self.previousMouseOffset:
                self.previousMouseOffset = offset
                if self.zoomDirection == -1:
                    self.zoomDirection = -1
                else:
                    self.zoomDirection = 1

            else:
                self.previousMouseOffset = offset
                self.zoomDirection = 1
                self.zoomIncr += 1

            if self.zoomDirection == 1:
                zoomFactor = 1.03
            else:
                zoomFactor = 1 / 1.03

            # Perform zoom and re-center on initial click position.
            pBefore = self.mapToScene(self.initMousePos)
            self.setTransformationAnchor(
                QtWidgets.QGraphicsView.AnchorViewCenter)
            self.scale(zoomFactor, zoomFactor)
            pAfter = self.mapToScene(self.initMousePos)
            diff = pAfter - pBefore

            self.setTransformationAnchor(QtWidgets.QGraphicsView.NoAnchor)
            self.translate(diff.x(), diff.y())

        # Drag canvas.
        elif self.currentState == 'DRAG_VIEW':
            offset = self.prevPos - event.pos()
            self.prevPos = event.pos()
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() + offset.y())
            self.horizontalScrollBar().setValue(
                self.horizontalScrollBar().value() + offset.x())

        # RuberBand selection.
        elif (self.currentState == 'SELECTION' or
              self.currentState == 'ADD_SELECTION' or
              self.currentState == 'SUBTRACT_SELECTION' or
              self.currentState == 'TOGGLE_SELECTION'):
            self.rubberband.setGeometry(QtCore.QRect(
                self.origin, event.pos()).normalized())

        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        # Zoom the View.
        if self.currentState == '.ZOOM_VIEW':
            self.offset = 0
            self.zoomDirection = 0
            self.zoomIncr = 0
            self.setInteractive(True)

        # Drag View.
        elif self.currentState == 'DRAG_VIEW':
            self.setCursor(QtCore.Qt.ArrowCursor)
            self.setInteractive(True)

        # Selection.
        elif self.currentState == 'SELECTION':
            self.rubberband.setGeometry(QtCore.QRect(self.origin,
                                                     event.pos()).normalized())
            painterPath = self._releaseRubberband()
            self.setInteractive(True)
            self.scene().setSelectionArea(painterPath)

        # Add Selection.
        elif self.currentState == 'ADD_SELECTION':
            self.rubberband.setGeometry(QtCore.QRect(self.origin,
                                                     event.pos()).normalized())
            painterPath = self._releaseRubberband()
            self.setInteractive(True)
            for item in self.scene().items(painterPath):
                item.setSelected(True)

        # Subtract Selection.
        elif self.currentState == 'SUBTRACT_SELECTION':
            self.rubberband.setGeometry(QtCore.QRect(self.origin,
                                                     event.pos()).normalized())
            painterPath = self._releaseRubberband()
            self.setInteractive(True)
            for item in self.scene().items(painterPath):
                item.setSelected(False)

        # Toggle Selection
        elif self.currentState == 'TOGGLE_SELECTION':
            self.rubberband.setGeometry(QtCore.QRect(self.origin,
                                                     event.pos()).normalized())
            painterPath = self._releaseRubberband()
            self.setInteractive(True)
            for item in self.scene().items(painterPath):
                if item.isSelected():
                    item.setSelected(False)
                else:
                    item.setSelected(True)

        self.currentState = 'DEFAULT'

        super().mouseReleaseEvent(event)

    def keyPressEvent(self, event):
        scene = self.scene()

        if event.key() not in self.pressedKeys:
            self.pressedKeys.append(event.key())

        if event.key() in (QtCore.Qt.Key_Delete, QtCore.Qt.Key_Backspace):
            self._deleteSelectedNodes()

        if event.key() == QtCore.Qt.Key_F:
            if event.modifiers() == QtCore.Qt.ShiftModifier:
                scene.setFormalPosition()
                scene.clearSelection()
            self._focus()

        items = scene.selectedItems()
        if items and (event.key() == QtCore.Qt.Key_R):
            for item in items:
                if isinstance(item, NodeItem):
                    self.setNodeAsRoot(item)

        # Emit signal.
        self.signal_KeyPressed.emit(event.key())

    def keyReleaseEvent(self, event):
        if event.key() in self.pressedKeys:
            self.pressedKeys.remove(event.key())

    def setNodeAsRoot(self, node):
        self.scene().root = node.data
        for connection in node.slot_parent.connections:
            connection._remove()
        node.update()

        # Emit signal
        self.signal_RootUpdated.emit(node)

    def _initRubberband(self, position):
        self.rubberBandStart = position
        self.origin = position
        self.rubberband.setGeometry(QtCore.QRect(self.origin, QtCore.QSize()))
        self.rubberband.show()

    def _releaseRubberband(self):
        painterPath = QtGui.QPainterPath()
        rect = self.mapToScene(self.rubberband.geometry())
        painterPath.addPolygon(rect)
        self.rubberband.hide()
        return painterPath

    def _focus(self, padding=50):
        if self.scene().selectedItems():
            itemsArea = self._getSelectionBoundingbox()
            self.fitInView(itemsArea, QtCore.Qt.KeepAspectRatio)
        else:
            itemsArea = self.scene().itemsBoundingRect()
            itemsArea.setX(itemsArea.x() - padding)
            itemsArea.setY(itemsArea.y() - padding -
                           self.config['root_font_size'])
            itemsArea.setWidth(itemsArea.width() + padding)
            itemsArea.setHeight(itemsArea.height() + padding)
            self.fitInView(itemsArea, QtCore.Qt.KeepAspectRatio)

    def _getSelectionBoundingbox(self):
        bbx_min = None
        bbx_max = None
        bby_min = None
        bby_max = None
        bbw = 0
        bbh = 0
        for item in self.scene().selectedItems():
            pos = item.scenePos()
            x = pos.x()
            y = pos.y()
            w = x + item.boundingRect().width()
            h = y + item.boundingRect().height()

            # bbx min
            if bbx_min is None:
                bbx_min = x
            elif x < bbx_min:
                bbx_min = x
            # end if

            # bbx max
            if bbx_max is None:
                bbx_max = w
            elif w > bbx_max:
                bbx_max = w
            # end if

            # bby min
            if bby_min is None:
                bby_min = y
            elif y < bby_min:
                bby_min = y
            # end if

            # bby max
            if bby_max is None:
                bby_max = h
            elif h > bby_max:
                bby_max = h
            # end if
        # end if
        bbw = bbx_max - bbx_min
        bbh = bby_max - bby_min
        return QtCore.QRectF(QtCore.QRect(bbx_min, bby_min, bbw, bbh))

    def _deleteSelectedNodes(self):
        selected_nodes = list()
        for node in self.scene().selectedItems():
            selected_nodes.append(node.name)
            node._remove()

        # Emit signal.
        self.signal_NodeDeleted.emit(selected_nodes)

    def _returnSelection(self):
        selected_nodes = list()
        if self.scene().selectedItems():
            for node in self.scene().selectedItems():
                selected_nodes.append(node.name)

        # Emit signal.
        self.signal_NodeSelected.emit(selected_nodes)

    def loadConfig(self, filePath):
        with filePath.open() as f:
            self.config = json.load(f)

    def initialize(self, root=None):
        # Setup view.
        config = self.config
        self.setRenderHint(
            QtGui.QPainter.Antialiasing,
            config['antialiasing']
        )
        self.setRenderHint(
            QtGui.QPainter.TextAntialiasing,
            config['antialiasing']
        )
        self.setRenderHint(
            QtGui.QPainter.HighQualityAntialiasing,
            config['antialiasing_boost']
        )
        self.setRenderHint(
            QtGui.QPainter.SmoothPixmapTransform,
            config['smooth_pixmap']
        )
        self.setRenderHint(QtGui.QPainter.NonCosmeticDefaultPen, True)
        self.setViewportUpdateMode(QtWidgets.QGraphicsView.FullViewportUpdate)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.rubberband = QtWidgets.QRubberBand(
            QtWidgets.QRubberBand.Rectangle,
            self
        )

        # Setup scene.
        scene = NodeScene(self)
        sceneWidth = config['scene_width']
        sceneHeight = config['scene_height']
        scene.setSceneRect(0, 0, sceneWidth, sceneHeight)
        self.setScene(scene)
        scene.setNodes(root)

        # Tablet zoom.
        self.previousMouseOffset = 0
        self.zoomDirection = 0
        self.zoomIncr = 0

        # Connect signals.
        self.scene().selectionChanged.connect(self._returnSelection)

    def createNode(self, data, position=None):
        nodeItem = NodeItem(self, data)

        # Store node in scene.
        self.scene().nodes[id(data)] = nodeItem

        if not position:
            # Get the center of the view.
            position = self.mapToScene(self.viewport().rect().center())

        # Set node position.
        self.scene().addItem(nodeItem)
        nodeItem.setPos(position - nodeItem.nodeCenter)
        nodeItem._createSlot()

        # Emit signal.
        self.signal_NodeCreated.emit(data.name)

        return nodeItem

    def deleteNode(self, node):
        if node in self.scene().nodes.values():
            nodeName = node.name
            node._remove()

            # Emit signal.
            self.signal_NodeDeleted.emit([nodeName])

    def createConnection(self, sourceData, targetData):
        id_source = id(sourceData)
        id_target = id(targetData)
        slot_child = self.scene().nodes[id_source].slot_child
        slot_parent = self.scene().nodes[id_target].slot_parent

        connection = ConnectionItem(
            slot_child.center(), slot_parent.center(),
            slot_child, slot_parent
        )

        connection.parentNode = slot_child.parentItem()
        connection.childNode = slot_parent.parentItem()

        slot_child.connect(slot_parent, connection)
        slot_parent.connect(slot_child, connection)

        connection.updatePath()

        self.scene().addItem(connection)

        return connection



import setuptools

setuptools.setup(
    name="qt-tree",
    version="0.1.0",
    author="Kotaro Kikuchi",
    author_email="ktrk115@gmail.com",
    description="Graphical user interface for editing and visualizing tree data structure",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ktrk115/qt-tree",
    packages=setuptools.find_packages(),
    install_requires=['PySide2', 'anytree'],
    extras_require={'pillow': ['pillow']},
    keywords='Qt, tree structure, tree editor',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
